/* trecov: recover unpreprocessed C code from unpreprocessed one
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: trecov-demo.c,v 1.1 2006/10/25 12:37:10 gondow Exp gondow $ */
/* ======================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/param.h> /* for PATH_MAX */

#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/tree.h>
#include <libxml/debugXML.h>

#include <curses.h>

#include "misc.h" /* for emalloc, erealloc */
#include "tef.h"

#define  DEFINE_LOCINFO                                        \
    xmlChar *filename_ref = xmlGetProp (node, "filename_ref"); \
    xmlChar *first_line   = xmlGetProp (node, "first_line");   \
    xmlChar *first_column = xmlGetProp (node, "first_column"); \
    xmlChar *last_line    = xmlGetProp (node, "last_line");    \
    xmlChar *last_column  = xmlGetProp (node, "last_column");
#define  FREE_LOCINFO                             \
    xmlFree (filename_ref);                       \
    xmlFree (first_line); xmlFree (first_column); \
    xmlFree (last_line);  xmlFree (last_column);

#ifndef MAX
#define MAX(a, b)	 ((a)>(b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a, b)	 ((a)>(b) ? (b) : (a))
#endif

enum {
    COLOR_BLACK2 = COLOR_WHITE + 1,
    COLOR_RED2, COLOR_GREEN2, COLOR_YELLOW2, 
    COLOR_BLUE2, COLOR_MAGENTA2, COLOR_CYAN2, COLOR_WHITE2,
};

/* ====================================================================== */
static void print_usage (int argc, char *argv []);
static int strlen_null (char *s);
static void trecov_init (int argc, char *argv[]);
static void trecov_final (void);
static void write_line_col_format (int row, int col, char *format, ...);
static void write_lines_col (char *str, int row, int col, int keep_trailer, int insert);
static void write_line_col (char *str, int row, int col, int keep_trailer, int insert);
static void dump_code_lines (int output_lineno);
static void dump_macrocall_node (xmlNodePtr node);
static char *get_line (FILE *fp);
static int get_file_id (xmlNodePtr node, char *attr_name);
static char *get_line (FILE *fp);
static xmlNodeSetPtr eval_xpath (char *xpath, xmlNodePtr from, void **free_p, int *num);
static xmlNodePtr get_child (xmlNodePtr parent, char *child);
static void set_file_id (void);
static void restore_macrocall (xmlNodePtr node, int *i, xmlNodeSetPtr result);
static xmlNodePtr next_macrocall_node (int *line, int *col);
static char *next_E_line (int *E_lineno);
static int  trim_line_col (char *str, int line, int col) ;
static void trecov_main (void);
static void mvprintw_with_scroll (int y, int x, char *fmt, char *str,
				  int color, int insert);
static void prompt (void);
static void curses_insert (int line, int col, char *str, int color);
/* ====================================================================== */
static xmlDocPtr  xml_doc  = NULL;
static xmlNodePtr xml_root = NULL;

static int  file_id = -1;
static char *filename;
static FILE *trecov_fp;
static FILE *E_fp;

/* code_lines is not NULL-terminated, while each code_lines[i] is
   '\0'-terminated. the line number starts from 1, not from 0. */
static char **code_lines;
static int  code_lines_size;
static int  max_lineno = 0;
/* ====================================================================== */
static void
print_usage (int argc, char *argv [])
{
    fprintf (stderr, "Usage: %s foo.c.E foo.c.xml\n", argv [0]);
}

static int
strlen_null (char *s)
{
    return (s == NULL ? 0 : strlen (s));
}

static void
trecov_init (int argc, char *argv[])
{
    char *p;

    code_lines_size = 2;
    code_lines = emalloc (sizeof (code_lines [0]) * code_lines_size);
    memset (code_lines, 0, sizeof (code_lines [0]) * code_lines_size);

    filename = strdup (argv [2]);
    p = strrchr (filename, '.');
    assert (p != NULL);
    *p = '\0';

    trecov_fp = stdout;

    E_fp = fopen (argv [1], "r");
    assert (E_fp != NULL);

    xmlDoValidityCheckingDefaultValue = 1;
    xmlSubstituteEntitiesDefault (1);
    xml_doc  = xmlParseFile(argv [2]);
    xml_root = xmlDocGetRootElement (xml_doc);
    xmlXPathInit();

    set_file_id ();

    /* init curses */
    initscr (); noecho (); cbreak (); scrollok (stdscr, TRUE);
#if 0
    idlok (stdscr, TRUE);
#endif
    if (has_colors ()) {
	start_color ();

	init_pair (COLOR_BLACK,   COLOR_BLACK,   COLOR_BLACK);
	init_pair (COLOR_RED,     COLOR_RED,     COLOR_BLACK);
	init_pair (COLOR_GREEN,   COLOR_GREEN,   COLOR_BLACK);
	init_pair (COLOR_YELLOW,  COLOR_YELLOW,  COLOR_BLACK);
	init_pair (COLOR_BLUE,    COLOR_BLUE,    COLOR_BLACK);
	init_pair (COLOR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
	init_pair (COLOR_CYAN,    COLOR_CYAN,    COLOR_BLACK);
	init_pair (COLOR_WHITE,   COLOR_WHITE,   COLOR_BLACK);

	init_pair (COLOR_BLACK2,   COLOR_BLACK,   COLOR_WHITE);
	init_pair (COLOR_RED2,     COLOR_RED,     COLOR_WHITE);
	init_pair (COLOR_GREEN2,   COLOR_GREEN,   COLOR_WHITE);
	init_pair (COLOR_YELLOW2,  COLOR_YELLOW,  COLOR_WHITE);
	init_pair (COLOR_BLUE2,    COLOR_BLUE,    COLOR_WHITE);
	init_pair (COLOR_MAGENTA2, COLOR_MAGENTA, COLOR_WHITE);
	init_pair (COLOR_CYAN2,    COLOR_CYAN,    COLOR_WHITE);
	init_pair (COLOR_WHITE2,   COLOR_WHITE,   COLOR_WHITE);
    }
}

static void
trecov_final (void)
{
    endwin (); /* finalize curses */
}

static void
write_line_col_format (int row, int col, char *format, ...)
{
    char *line;
    int size;
    va_list pvar;

    va_start (pvar, format);
    size = vsnprintf (NULL, 0, format, pvar);
    va_end (pvar);

    line = emalloc (size + 1); /* +1 for '\0' */
    va_start (pvar, format);
    vsnprintf (line, size + 1, format, pvar);
    va_end (pvar);

    write_line_col (line, row, col, 0, 0);
    free (line);
}

static void
write_lines_col (char *str, int row, int col, int keep_trailer, int insert)
{
    char *str2 = estrdup (str);
    char *p = str2, *q;
    int curr_row = row;

    while (1) {
	q = strchr (p, '\n');
	if (q == NULL) {
	    write_line_col (p, curr_row, (curr_row == row ? col: 0),
			    keep_trailer, insert);
	    break;
	} else {
	    *q = '\0';
	    write_line_col (p, curr_row, (curr_row == row ? col: 0),
			    0, 0);
	}
	curr_row++;
	p = q + 1;
    }
    EFREE (str2);
}

static void
write_line_col (char *str, int row, int col, int keep_trailer, int insert)
{
    /* col==0 means the beginning of line; -1 means the end of line*/
    int str_len;
    int line_len;

    assert (col >= -1);

    if (row > max_lineno)
	max_lineno = row;

    /* enough memory allocated for code_lines? */
    while (row >= code_lines_size) {
	code_lines_size *= 2;
	code_lines = erealloc (code_lines, 
			       sizeof (code_lines [0]) * code_lines_size);
	memset (code_lines + code_lines_size / 2, 0,
		sizeof (code_lines [0]) * (code_lines_size / 2));
    }

    assert (0 <= row  && row < code_lines_size);

    if (code_lines [row] == NULL) {
	code_lines [row] = estrdup ("");
    }

    line_len = strlen_null (code_lines [row]);
    str_len  = strlen_null (str);

    assert (code_lines [row][line_len] == '\0');
    assert (str [str_len] == '\0');  /* str must be '\0'-terminated */

    if (col == -1) col = line_len;

    /* enough memory allocated for the line? */
    if (insert) {
	code_lines [row] = erealloc (code_lines [row],
				     MAX (line_len, col) + str_len + 1);
    } else if (line_len < col + str_len) {
	code_lines [row] = erealloc (code_lines [row], col + str_len + 1);
    }

    /* now data transfered */
    if (insert) {
	if (line_len >= col)
	    memmove (code_lines [row] + col + str_len,
		     code_lines [row] + col, line_len - col + 1);
	memcpy (code_lines [row] + col, str, str_len);
    } else if (line_len >= col + str_len && keep_trailer) {
	/* preserving the already existing line */
	memcpy (code_lines [row] + col, str, str_len);
    } else {
	memcpy (code_lines [row] + col, str, str_len + 1);
#if 1
	move (row, col + str_len); clrtoeol ();
#endif
    }

    /* fill the hole with space characters */
    if (line_len < col) {
	memset (code_lines [row] + line_len, ' ', col - line_len);
    }
}

static void
dump_code_lines (int output_lineno)
{
    int i;
    char *dump_filename;
    FILE *fp;

    easprintf (&dump_filename, "%s.trecov", filename);
    fp = fopen (dump_filename, "w");
    if (fp == NULL) {
	perror ("fopen"); exit (1);
    }
    
    for (i = 1; i < max_lineno; i++) {
	assert (i < code_lines_size);
	if (output_lineno)
	    fprintf (fp, "%5d: ", i);
	if (code_lines [i] == NULL) {
	    fprintf (fp, "\n");
	} else {
	    fprintf (fp, "%s\n", code_lines [i]);
	}
    }
    fclose (fp);
    EFREE (dump_filename);
}

static void
dump_macrocall_node (xmlNodePtr node)
{
    DEFINE_LOCINFO
    xmlChar *type = xmlGetProp (node, "type");
    xmlChar *name = xmlGetProp (node, "name");
    xmlChar *args = xmlGetProp (node, "args");
    xmlChar *bare = xmlNodeGetContent (get_child (node, "bare"));

    fprintf (stderr, "type=%s, name=%s, args=%s, bare=%s\n",
	     type, name, args, bare);
    fprintf (stderr, "first_line=%s, first_column=%s, last_line=%s, last_column=%s\n", first_line, first_column, last_line, last_column);
    FREE_LOCINFO
    xmlFree (type); xmlFree (name); xmlFree (args); xmlFree (bare);
}

static char *
get_line (FILE *fp)
{
    static char *buf  = NULL;
    static int   size = 0;
    char *p;

    if (buf == NULL) {
	assert (size == 0);
	size = 1024;
	buf = emalloc (size);
    }

    p = buf;
    
    while (1) {
        fgets (p, size - (p - buf), fp);

	if (feof (fp))
	    return NULL;
	if (ferror (fp)) {
	    perror ("fgets");
	    exit (1);
	}
	if (buf [strlen (buf) - 1] == '\n') { /* entire line obtained */
	    buf [strlen (buf) - 1] = '\0'; /* chop the newline */
	    return buf;
	}

	size *= 2;
	buf = erealloc (buf, size);
	p = buf + size / 2 - 1;
    }
    /* not reached */
    return 0;
}

static int
get_file_id (xmlNodePtr node, char *attr_name)
{
    xmlChar *id;
    char *p;
    int id_num;

    id = xmlGetProp (node, attr_name);
    if (id == NULL)
	return -2;

    assert (id [0] == 'F');
    p = strchr (id, '_');
    if (p != NULL)
	*p = '\0';

    id_num = atoi (id + 1);
    xmlFree (id);
    return id_num;
}

static xmlNodeSetPtr 
eval_xpath (char *xpath, xmlNodePtr from, void **free_p, int *num)
{
    xmlXPathObjectPtr  result;
    xmlXPathContextPtr context;

    context = xmlXPathNewContext(xml_doc);
    assert (context != NULL);
    context->node = from;
    result = xmlXPathEval (xpath, context);
    xmlXPathFreeContext (context);
#if 0
    xmlXPathDebugDumpObject (stdout, result, 0);
#endif
    if (num != NULL)
	*num = xmlXPathNodeSetGetLength (result->nodesetval);
    if (free_p != NULL)
	*free_p = result;
    return result->nodesetval;
}

static xmlNodePtr
get_child (xmlNodePtr parent, char *child)
{
    xmlNodePtr node;

    if (parent->children == NULL)
	return NULL;

    for (node = parent->children; node != NULL; node = node->next) {
	if (node->type != XML_ELEMENT_NODE) continue;
	if (strcmp (node->name, child))     continue;
	return node;
    }
    return NULL;
}

static void
set_file_id (void)
{
    xmlNodeSetPtr result;
    int n;
    void *free_p;
    xmlNodePtr  node;

    result = eval_xpath ("/TBCppA/file[1]", xml_root, &free_p, &n);
    node = result->nodeTab [0];
    assert (n == 1);
    assert (!strcmp(node->name, "file"));
    file_id = get_file_id (node, "id");
    xmlFree (free_p);
}

static void 
restore_macrocall (xmlNodePtr node, int *i, xmlNodeSetPtr result)
{
    /* obj, func, param, defined */
    DEFINE_LOCINFO
    xmlChar *type = xmlGetProp (node, "type");
    xmlChar *name = xmlGetProp (node, "name");
    xmlChar *args = xmlGetProp (node, "args");
    xmlChar *bare = xmlNodeGetContent (get_child (node, "bare"));
    char *new, *p;
    int n;
    int line = atoi (first_line);
    int col  = atoi (first_column);

    assert (strcmp (type, "param") && strcmp (type, "defined"));

    if (bare == NULL) bare = xmlStrdup ("");
    if (args == NULL)
	new = estrdup (name);
    else
	easprintf (&new, "%s(%s)", name, args);

    n = trim_line_col (new, line, col);
    while (n--)
	mvdelch (line, col);
    write_lines_col (new, line, col, 0, 1);
    mvprintw_with_scroll (line, col, "%s", new, COLOR_RED, 0);
    refresh ();
    prompt ();

    p = new;
    while (1) {
	xmlNodePtr lex_node = result->nodeTab[*i];
	char *s = lex_node->children->content;
	int len_p = strlen (p);
	int len_s = strlen (s);
	int len_min = (len_p > len_s ? len_s : len_p);
	assert (s != NULL);
	if (len_p == 0 || strncmp (p, s, len_min)) {
	    (*i)--;
	    break;
	}
	p += strlen (s);
	(*i)++;
    }

    FREE_LOCINFO
    xmlFree (type); xmlFree (name); xmlFree (args); xmlFree (bare);
    EFREE (new);
}

static xmlNodePtr
next_macrocall_node (int *line, int *col)
{
    static xmlNodeSetPtr result = NULL;
    static int n, i = 0;

    assert (file_id != -1);
    if (result == NULL) {
	char *xpath;

	easprintf (&xpath, "//macro_call[not(ancestor::macro_call)"
		   "and not(../../../cond_selected) and @filename_ref='F%d' "
		   "and @type!='defined' and @type!='param']", file_id);
	result = eval_xpath (xpath, xml_root, NULL, &n);
	free (xpath);
    }

    if (i < n) {
	xmlNodePtr node = result->nodeTab[i];
	DEFINE_LOCINFO
	assert (get_file_id (node, "filename_ref") == file_id);
	if (line != NULL) *line = atoi (first_line);
	if (col  != NULL) *col  = atoi (first_column);
	FREE_LOCINFO
	i++;
	return node;
    }

    if (line != NULL) *line = -1;
    if (col  != NULL) *col  = -1;
    return NULL;
}

static char *
next_E_line (int *E_lineno)
{
    static char *path = NULL;
    static int lineno = 0, do_print = 0;
    int n, n_match;

    if (path == NULL) {
	int path_max;
#ifdef PATH_MAX
	path_max = PATH_MAX;
#else
	path_max = getconf (".", _PC_PATH_MAX);
#endif
	if (path_max < 1024)
	    path_max = 1024;
	path = emalloc (path_max + 1);
    }

    while (1) {
	char *line = get_line (E_fp);
	if (line == NULL)
	    return NULL;
	n_match = sscanf (line, "# %d \"%[^\"]", &n, path);
	if (n_match == 2) {
	    if (!strcmp (path, filename)) {
		lineno = n;
		do_print = 1;
	    } else {
		do_print = 0;
	    }
	} else if (do_print) {
	    if (E_lineno != NULL)
		*E_lineno = lineno;
	    lineno++;
	    return line;
	}
    }
}

static int
trim_line_col (char *str, int line, int col) 
{
    int ret = 0;
    char *str2;
    char *p;

    if (code_lines [line] == NULL)
	return ret;

    str2 = estrdup (str);
    p = strchr (str2, '\n');
    if (p != NULL) p = '\0';

    p = strstr (code_lines [line] + col, str2);
    if (p != NULL && code_lines [line] + col != p) {
	memmove (code_lines [line] + col, p, strlen (p) + 1);
	ret = p - (code_lines [line] + col);
    }

    EFREE (str2);
    return ret;
}

static
void trecov_main (void)
{
    int mc_line, mc_col, pp_line, pp_col, pp_line2;
    xmlNodePtr macrocall_node;
    xmlNodeSetPtr result;
    int n, i;
    void *free_p;
    char *E_line_str;
    int E_line;

    macrocall_node = next_macrocall_node (&mc_line, &mc_col);
    E_line_str = next_E_line (&E_line);

    result = eval_xpath ("/TBCppA/lexical_info/*", xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr node = result->nodeTab [i];
	DEFINE_LOCINFO
	xmlChar *type = xmlGetProp (node, "type");

	pp_line  = atoi (first_line);
	pp_col   = atoi (first_column);
	pp_line2 = atoi (last_line);

	while (E_line_str != NULL && E_line <= pp_line2) {
	    write_line_col (E_line_str, E_line, 0, 0, 0);
	    mvprintw_with_scroll (E_line, 0, "%s", E_line_str,
				  COLOR_MAGENTA2, 0);
	    refresh ();
	    prompt ();
	    E_line_str = next_E_line (&E_line);
	}

	if (!strcmp (node->name, "pp_token")) {
	    if (mc_line == pp_line && mc_col == pp_col) {
		assert (macrocall_node != NULL);
		restore_macrocall (macrocall_node, &i, result);
		macrocall_node = next_macrocall_node (&mc_line, &mc_col);
	    } else {
		int n;
		n = trim_line_col (node->children->content, pp_line, pp_col);
		while (n--)
		    mvdelch (pp_line, pp_col);
		write_lines_col (node->children->content,
				 pp_line, pp_col, 1, 0);
		mvprintw_with_scroll (pp_line, pp_col, "%s",
				      node->children->content,
				      COLOR_WHITE, 0);
		refresh ();
		prompt ();
	    }
	} else if (!strcmp (node->name, "comment")) {
	    write_lines_col (node->children->content, pp_line, pp_col, 0, 1);
	    mvprintw_with_scroll (pp_line, pp_col, "%s",
				  node->children->content, COLOR_WHITE, 1);
	    refresh ();
	} else if (!strcmp (node->name, "white_space")) {
	    /* special handling of formfeed characters (^L, &#x0C;) */
	    char *ws = xmlNodeGetContent (node);
	    int  ff_len = strlen ("&#x0C;");
	    while (1) {
		char *p = strstr (ws, "&#x0C;");
		if (p == NULL) break;
		*p = '\f';
		memmove (p + 1, p + ff_len, strlen (p) - ff_len + 1);
	    }
	    write_lines_col (ws, pp_line, pp_col, 0, 1);
	    mvprintw_with_scroll (pp_line, pp_col, "%s", ws, COLOR_WHITE, 1);
	    EFREE (ws);
	    refresh ();
	} else {
	    assert (0);
	}

	FREE_LOCINFO
        xmlFree (type);
    }
    
    free (free_p);
}

static void
mvprintw_with_scroll (int y, int x, char *fmt, char *str, int color,
		      int insert)
{
    static int line = 0;

    if (y < line || line + LINES <= y) {
	int j;
	scrl (y - line + LINES / 2);
	line = y - LINES / 2;
	for (j = 0; j < LINES / 2 + 1; j++) {
	    char *str2 = code_lines [y - (LINES / 2 - j)];
	    mvprintw (j, 0, "%s", (str2 == NULL ? "" : str2));
	}
    }
    if (insert) {
	curses_insert (y-line, x, (str == NULL ? "" : str), color);
    } else {
	attron (COLOR_PAIR (color));
	attron (A_BOLD);
	mvprintw (y - line, x, fmt, (str == NULL ? "" : str));
    }
    refresh ();
}

static void
prompt (void)
{
    static int is_auto = 1;
    chtype ch;

    ch = getch ();

    if (ch == ERR)
	return;

    if ((ch & A_CHARTEXT) == 'a') {
	is_auto = !is_auto;
	if (is_auto)
	    nodelay (stdscr, TRUE);
	else
	    nodelay (stdscr, FALSE);
    }
}

static void
curses_insert (int line, int col, char *str, int color)
{
    enum { NCHAR = 2048 };
    chtype buf [NCHAR];

    mvinchnstr (line, col, buf, NCHAR);
    mvaddchstr (line, col + strlen (str), buf);
    attron (COLOR_PAIR (color));
    attron (A_BOLD);
    mvprintw (line, col, str);
    prompt ();
}
/* ====================================================================== */
int
main (int argc, char *argv [])
{
#if defined(USE_TEF)
    extern int TEF_is_in_main;
    TEF_is_in_main = 1;
#endif

    if (argc < 2) {
	print_usage (argc, argv);
	exit (1);
    }

    trecov_init (argc, argv);
    trecov_main ();
    dump_code_lines (0);
    trecov_final ();
    return 0;
}

#if 0
Efoo.c.E Ƃ̃}b`O (ǂ݂邩?)
Ecurses gāCAj[V\ɂD

insert ̂ƂɁC\}łĂȂD

curses_insert ł͈s̍ő啶ŒŎwD
#endif
